; RC5
; assembly implementation (x) herm1t@vx.netlux.org 2006-08-19
;	void rc5_encrypt(uint32_t *key, uint32_t *data, int blocks[, int rounds]);
;	void rc5_decrypt(uint32_t *key, uint32_t *data, int blocks[, int rounds]);
;	void rc5_key(uint32_t *key, uint8_t *userkey, int keylen[, int rounds]);
; with -DRARG option each function will accept additional argument which
; specify the number of rounds
;
; Bruce Schneier "Applied Cryptography, Second Edition: Protocols, Algorthms,
;	and Source Code in C", John Wiley & Sons, Inc., 1996, ISBN: 0471128457
;
; 14.8 RC5
;    RC5 is a block cipher with a variety of parameters: block size, key size,
; and number of rounds. It was invented by Ron Rivest and analyzed by RSA
; Laboratories [1324,1325].
;    There are three operations: XOR, addition, and rotations. Rotations are
; constant-time operations on most processors and variable rotations are a
; nonlinear function. These rotations, which depend on both the key and the
; data, are the interesting operation.
; RC5 has a variable-length block, but this example will focus on a 64-bit
; data block. Encryption uses 2r + 2 key-dependent 32-bit words S0,S1,S2,...,
; S2r + 1 where r is the number of rounds. Well generate those words later.
; To encrypt, first divide the plaintext block into two 32-bit words: A and B.
; (RC5 assumes a little-endian convention for packing bytes into words: The
; first byte goes into the low-order bit positions of register A, etc.) Then:
;	A = A + S0 
;	B = B + S1 
;	For i = 1 to r:
;		A = ((A ^ B) <<< B) + S2i 
;		B = ((B ^ A) <<< A) + S2i + 1 
; The output is in the registers A and B.
; Decryption is just as easy. Divide the plaintext block into two words, A and
; B, and then:
; 	For i = r down to 1:
;		B = ((B ^ S2i + 1) >>> A) ^ A
;		A = ((A ^ S2i) >>> B) ^ B
; 	B = B - S1 
;	A = A - S0 
; The symbol >>> is a right circular shift. Of course, all addition and
; subtraction are mod 2^32.
;    Creating the array of keys is more complicated, but also straightforward.
; First, copy the bytes of the key into an array, L, of c 32-bit words, padding
; the final word with zeros if necessary. Then, initialize an array, S, using a
; linear congruential generator mod 2^32:
;	S0 = P 
;	for i = 1 to 2(r + 1)  1:
;		Si = (Si - 1 + Q) mod 2^32 
; P = 0xb7e15163 and Q = 0x9e3779b9; these constants are based on the binary
; representation of e and phi.
; Finally, mix L into S:
;	i = j = 0 
;	A = B = 0 
; do 3n times (where n is the maximum of 2(r + 1) and c):
;	A = Si = (Si + A + B) <<< 3 
;	B = Lj = (Lj + A + B) <<< (A + B) 
;	i = (i + 1) mod 2(r + 1) 
;	j = (j + 1) mod c 
;   RC5 is actually a family of algorithms. We just defined RC5 with a 32-bit
; word size and 64-bit block; theres no reason why the same algorithm cant
; have a 64-bit word size and 128-bit block size. For w = 64, P and Q are
; 0xb7e151628aed2a6b and 0x9e3779b97f4a7c15, respectively. Rivest designates
; particular implementations of RC5 as RC5-w/r/b, where w is the word size, r
; is the number of rounds, and b is the length of the key in bytes.
;    RC5 is new, but RSA Laboratories has spent considerable time analyzing it
; with a 64-bit block. After 5 rounds, the statistics look very good. After 8
; rounds, every plaintext bit affects at least one rotation. There is a
; differential attack that requires 2^24 chosen plaintexts for 5 rounds, 2^45
; for 10 rounds, 2^53 for 12 rounds, and 2^68 for 15 rounds. Of course, there
; are only 2^64 possible chosen plaintexts, so this attack wont work for 15 or
; more rounds. Linear cryptanalysis estimates indicate that it is secure after
; 6 rounds. Rivest recommends at least 12 rounds, and possibly 16 [1325]. This
; number may change.
;    RSADSI is in the process of patenting RC5, and the name is trademarked.
; The company claims that license fees will be very small, but youd better
; check with them.
	    BITS	32
	    global	rc5_encrypt
	    global	rc5_decrypt
	    global	rc5_key
%define	ROUNDS	12
%macro _mov 2
%if %2 == 0
	xor	%1, %1
%elif %2 == 1
	xor	%1, %1
	inc	%1
%elif %2 < 128
	push	%2
	pop	%1
%else
	mov	%1, %2
%endif
%endmacro

rc5_encrypt:	pusha
		mov	edi, [esp + 36]		; key
		mov	esi, [esp + 40]		; data
		mov	ecx, [esp + 44]		; blocks
%ifdef	RARG
		mov	edx, [esp + 48]		; rounds
		shl	edx, 1
%endif
	.l0:	mov	eax, [edi + 0]
		add	[esi + 0], eax
		mov	eax, [edi + 4]
		add	[esi + 4], eax
		
		xor	ebp, ebp
		
	.l1:		push	ecx
			mov	eax, [esi + 0]
			mov	ebx, [esi + 4]

    			    xor	eax, ebx
			    mov	ecx, ebx
			    and	ecx, 31
			    rol	eax, cl
			    add	eax, [edi + ebp * 4 + 8]

			    xor	ebx, eax
			    mov	ecx, eax
			    and	ecx, 31
			    rol	ebx, cl
			    add	ebx, [edi + ebp * 4 + 12]

		mov	[esi + 0], eax
		mov	[esi + 4], ebx
		pop	ecx
	
		inc	ebp
		inc	ebp
%ifdef	RARG
		cmp	ebp, edx
%else
		cmp	ebp, byte (ROUNDS * 2)
%endif
		jb	.l1

		lea	esi, [esi + 8]
		loop	.l0
		popa
		ret

rc5_decrypt:	pusha
		mov	edi, [esp + 36]		; key
		mov	esi, [esp + 40]		; data
		mov	ecx, [esp + 44]		; blocks
%ifdef	RARG
		mov	edx, [esp + 48]		; rounds
		shl	edx, 1
		dec	edx
		dec	edx
	.l0:	mov	ebp, edx
%else
	.l0:	_mov	ebp, (ROUNDS * 2 - 2)
%endif
	.l1:		push	ecx
			mov	eax, [esi + 0]
			mov	ebx, [esi + 4]
			
			sub	ebx, [edi + ebp * 4 + 12]
			mov	ecx, eax
			and	cl, 31
			ror	ebx, cl
			xor	ebx, eax
			sub	eax, [edi + ebp * 4 + 8]
			mov	ecx, ebx
			and	cl, 31
			ror	eax, cl
			xor	eax, ebx
			
			mov	[esi + 0], eax
			mov	[esi + 4], ebx
			
			pop	ecx
		dec	ebp
		dec	ebp
		jns	.l1
		
		mov	eax, [edi + 4]
		sub	[esi + 4], eax
		mov	eax, [edi + 0]
		sub	[esi + 0], eax	
		lea	esi, [esi + 8]
		loop	.l0
		
		popa
		ret

rc5_key:	pusha

;	int	pk_len;
;	pk_len = keylen/4;
;	if ((keylen%4)!=0)
;	    pk_len += 1;
		mov	ecx, [esp + 44]		; keylen
		mov	edx, ecx
		shr	ecx, 2
		test	dl, 3
		jz	.l0
		inc	ecx	
.l0:		mov	ebx, ecx		; save pk_len (dwords)
		shl	ecx, 2

;	int	i;
;	u4	pk[pk_len];
;	for (i=0; i<pk_len; i++)
;	    pk[i]=0;
;       for (i=0; i<keylen; i++)
;    	    ((u1*)pk)[i] = key[i];

		; ecx - pk_len (bytes)
		; edx - keylen
		mov	ebp, esp
		xor	eax, eax
		push	eax			; A
		push	eax			; B
		sub	esp, ecx

		cld
		mov	edi, esp
		rep	stosb
		mov	ecx, edx
		mov	esi, [ebp + 40]		; userkey
		mov	edi, esp
		rep	movsb
	
;	int	xk_len = rounds*2 + 2;
;	xk[0] = 0xb7e15163;
;	for (i=1;i<xk_len; i++)
;		xk[i] = xk[i-1] + 0x9e3779b9;	
		mov	ecx, ebx		; restore pk_len (dwords)
		mov	esi, [ebp + 36]		; key
		mov	dword [esi], 0xb7e15163	; P32
%ifdef	RARG
		mov	edx, [ebp + 48]		; rounds
		shl	edx, 1
		inc	edx
		inc	edx
		_mov	eax, 0
.l1:		mov	ebx, [esi + eax*4]
		add	ebx, 0x9e3779b9		; Q32
		mov	[esi + eax * 4 + 4], ebx
		inc	eax
		cmp	eax, edx
		jb	.l1		
%else
		_mov	edx, (ROUNDS * 2 + 2)
		pusha
		mov	ecx, edx		
	.l1:	lodsd
		add	eax, 0x9e3779b9
		mov	edi, esi		
		stosd
		loop	.l1
		popa
%endif
		mov	edi, esp
		; ecx - pk_len (dwords)
		; edx - xk_len
		; esi - xk
		; edi - pk
		; [ebp - 4] - A
		; [ebp - 8] - B

;	u4	A = 0, B = 0;
;	for (i=0; i < max(pk_len, xk_len) * 3; i++) {
;		xk[i%xk_len] = ROTL32(xk[i%xk_len] + A + B, 3);
;		A = xk[i%xk_len];
;		int rc = (A + B) & 31;
;		pk[i%pk_len] = ROTL32(pk[i%pk_len] + A + B, rc);
;		B = pk[i%pk_len];
;	}

		mov	ebx, ecx
		cmp	ebx, edx
		ja	.c1
		mov	ebx, edx
.c1:		lea	ebx, [ebx * 2 + ebx]
		_mov	eax, 0

.l2:		pusha
		mov	ebx, edx
		cdq
		div	ebx
		mov	eax, [esi + edx * 4]
		add	eax, [ebp - 4]
		add	eax, [ebp - 8]
		rol	eax, 3
		mov	[esi + edx * 4], eax
		mov	[ebp - 4], eax
		popa
		
		pusha
		cdq
		div	ecx
		mov	eax, [edi + edx * 4]
		mov	ecx, [ebp - 4]
		add	ecx, [ebp - 8]
		add	eax, ecx
		and	cl, 31
		rol	eax, cl
		mov	[edi + edx * 4], eax
		mov	[ebp - 8], eax
		popa

		inc	eax
		cmp	eax, ebx
		jb	.l2

		mov	esp, ebp
		popa
		ret
